#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "ap_config.h"

#include <fcntl.h>
#include <sys/types.h>
#include <unistd.h>

#include "blosxom.h"

static const char HTML_CONTENT_TYPE[] = "text/html";
static const char HTML_HEAD[] = "<html><head><link rel=\"alternate\" type=\"type=\"application/rss+xml\" title=\"RSS\" href=\"$url/?flav=rss\" /><title>$blog_title $path_info_da $path_info_mo $path_info_yr</title></head><body><center><font size=\"+3\">$blog_title</font><br />$path_info_da $path_info_mo $path_info_yr</center><p />";
static const char HTML_STORY[] = "<p><a name=\"$fn\"><b>$title</b></a><br />$body<br /><br />posted at: $ti | path: <a href=\"$url$path\">$path</a> | <a href=\"$url/$yr/$mo_num/$da#$fn\">permanent link to this entry</a></p>\n";
static const char HTML_DATE[] = "<h3>$dw, $da $mo $yr</h3>\n";
static const char HTML_FOOT[] = "<p /><center><a href=\"http://www.raelity.org/apps/blosxom/\"><img src=\"http://www.raelity.org/images/pb_blosxom.gif\" border=\"0\" /></body></html>";

static const char RSS_CONTENT_TYPE[] = "text/xml";
static const char RSS_HEAD[] = "<?xml version=\"1.0\"?>\n<!-- name=\"generator\" content=\"blosxom/$version\" -->\n<!DOCTYPE rss PUBLIC \"-//Netscape Communications//DTD RSS 0.91//EN\" \"http://my.netscape.com/publish/formats/rss-0.91.dtd\">\n\n<rss version=\"0.91\">\n  <channel>\n    <title>$blog_title $path_info_da $path_info_mo $path_info_yr</title>\n    <link>$url</link>\n    <description>$blog_description</description>\n    <language>$blog_language</language>\n";
static const char RSS_STORY[] = "  <item>\n    <title>$title</title>\n    <link>$url/$yr/$mo_num/$da#$fn</link>\n    <description>$body</description>\n  </item>\n";
static const char RSS_DATE[] = "\n";
static const char RSS_FOOT[] = "  </channel>\n</rss>";

static const char ERROR_CONTENT_TYPE[] = "text/html";
static const char ERROR_HEAD[] = "<html><body><p><font color=\"red\">Error: I'm afraid this is the first I've heard of a \"$flavour\" flavoured Blosxom.  Try dropping the \"/+$flavour\" bit from the end of the URL.</font>\n\n";
static const char ERROR_STORY[] = "<p><b>$title</b><br />$body <a href=\"$url/$yr/$mo_num/$da#fn.$default_flavour\">#</a></p>\n";
static const char ERROR_DATE[] = "<h3>$dw, $da $mo $yr</h3>\n";
static const char ERROR_FOOT[] = "</body></html>";

static template_rec TEMPLATE[] = {
    { "html",  "content_type", "text/html" },
    { "html",  "head", HTML_HEAD },
    { "html",  "story", HTML_STORY },
    { "html",  "date", HTML_DATE },
    { "html",  "foot", HTML_FOOT },
    { "rss",   "content_type", "text/xml" },
    { "rss",   "head", RSS_HEAD },
    { "rss",   "story", RSS_STORY },
    { "rss",   "date", RSS_DATE },
    { "rss",   "foot", RSS_FOOT },
    { "error", "content_type", "text/html" },
    { "error", "head", ERROR_HEAD },
    { "error", "story", ERROR_STORY },
    { "error", "date", ERROR_DATE },
    { "error", "foot", ERROR_FOOT },
    { NULL, NULL, NULL }
};


static const char *mb_core_template(blosxom_context_rec *ctx, const char *chunk)
{
    char *path;
    struct stat f_info;
    int fd;
    char *body;
    int i;

    path = ctx->path_info;
    while (*path && (path = strchr(path, '/')) != NULL) {
        char *template;
        ssize_t rs;

        path++;
        if (stat(template, &f_info) < 0)
            continue;
        if ((fd = ap_popenf(ctx->pool, template, O_RDONLY, 0644)) < 0) {
            ap_log_rerror(APLOG_MARK, APLOG_WARNING, ctx->request_rec,
                          "cannot open template: %s", template);
            continue;
        }
        body = ap_palloc(ctx->pool, f_info.st_size + 1);
        rs = read(fd, (void *)body, f_info.st_size);
        if (rs < f_info.st_size) {
            ap_log_rerror(APLOG_MARK, APLOG_WARNING, ctx->request_rec,
                          "cannot read all content: %s", template);
        }
        body[f_info.st_size + 1] = '\0';
        ap_pclosef(ctx->pool, fd);
        return body;
    }

    for (i = 0; TEMPLATE[i].flavour != NULL; i++) {
        if (strcmp(ctx->flavour, TEMPLATE[i].flavour) != 0)
            continue;
        if (strcmp(chunk, TEMPLATE[i].chunk) != 0)
            continue;
        return TEMPLATE[i].body;
    }

    for (i = 0; TEMPLATE[i].flavour != NULL; i++) {
        if (strcmp("error", TEMPLATE[i].flavour) != 0)
            continue;
        if (strcmp(chunk, TEMPLATE[i].chunk) != 0)
            continue;
        return TEMPLATE[i].body;
    }
   
    ap_log_rerror(APLOG_MARK, APLOG_WARNING|APLOG_NOERRNO, ctx->request_rec,
                  "cannot read default template: flavour=%s chunk=%s",
                  "error", chunk);
    return "";
}

static int is_match_file_extension(char *fname, const char *file_extension)
{
    char *ext;

    if ((ext = strrchr(fname, '.')) == NULL)
        return 0;
    return strcmp(ext + 1, file_extension) == 0; 

}

static int find_entries(blosxom_context_rec *ctx, const char *path, array_header *ent)
{
    blosxom_conf_rec *conf;
    DIR *dh;
    struct dirent *node;
    char *f_path;
    struct stat finfo;
    blosxom_entry *new_entry;

    conf = ap_get_module_config(ctx->request_rec->per_dir_config,
                                &blosxom_module);
    if ((dh = ap_popendir(ctx->pool, path)) == NULL) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, ctx->request_rec,
                      "cannot open blosxom entry dir: %s", path);
        return DECLINED;
    }
    while ((node = readdir(dh)) != NULL) {
        if ((node->d_name)[0] == '.')
            continue;
	f_path = ap_make_full_path(ctx->pool, path, node->d_name);
        ap_no2slash(f_path);
        if (stat(f_path, &finfo) == -1)
            continue;
        if (S_ISDIR(finfo.st_mode)) {
            find_entries(ctx, f_path, ent);
            continue;
        }
        if (!S_ISREG(finfo.st_mode))
            continue;
        if (!is_match_file_extension(f_path, conf->file_extension))
            continue;

        new_entry = ap_pcalloc(ctx->pool, sizeof(blosxom_entry));
        new_entry->source = f_path;
        new_entry->mtime = finfo.st_mtime;
        *(blosxom_entry **)ap_push_array(ent) = new_entry;
    }
    ap_pclosedir(ctx->pool, dh);

    return OK;
}

static int mb_core_entries(blosxom_context_rec *ctx)
{
    blosxom_conf_rec *conf;
    array_header *entries;
    char *path;

    conf = ap_get_module_config(ctx->request_rec->per_dir_config,
                                &blosxom_module);
    path = ap_make_full_path(ctx->pool, conf->datadir, ctx->path_info);
    ap_no2slash(path);
    entries = ap_make_array(ctx->pool, 10, sizeof(blosxom_entry *));
    find_entries(ctx, path, entries);
    ctx->entries = entries;

    return OK;
}

static const char *mb_core_interpolate(blosxom_context_rec *ctx,
                                       const char *text)
{
    char *pos, *head;
    int max = strlen(text);
    char *new_text = "";

    for (pos = (char *)text; *pos != '\0'; pos++) {
        char *key, *value, *last_pos;
	if (max < (pos + 2) - text) {
            new_text = ap_pstrcat(ctx->pool, new_text, pos, NULL);
            break;
        }
        if (*pos == '$') {
            pos++;
            head = pos;
            for (; *pos != '\0' && (ap_isalpha(*pos) || *pos == '_'); pos++)
                ;
            key = ap_pstrndup(ctx->pool, head, pos - head);
            key[pos - head] = '\0';
            value = (char *)ap_table_get(ctx->bind_string, key);
            new_text = ap_pstrcat(ctx->pool,
                                  new_text, (value ? value : ""), NULL);
            pos--;
            continue;
	}

        last_pos = pos;
        for (; *pos && *pos != '$'; pos++)
            ;
        new_text = ap_pstrcat(ctx->pool, new_text,
                              ap_pstrndup(ctx->pool, last_pos, pos - last_pos),
                              NULL);
        pos--;
    }

    return new_text;
}

static int mb_core_sort_cmp(const void *a, const void *b)
{
    blosxom_entry *e1 = *(blosxom_entry **)a;
    blosxom_entry *e2 = *(blosxom_entry **)b;
    if (e1->mtime < e2->mtime)
        return 1;
    if (e1->mtime > e2->mtime)
        return -1;
    else
        return 0;
}

static int mb_core_sort(blosxom_context_rec *ctx)
{
    qsort(ctx->entries->elts, ctx->entries->nelts,
          sizeof(blosxom_entry *), mb_core_sort_cmp);
    return OK;
}

blosxom_plugin core_blosxom_plugin = {
    BLOSXOM_PLUGIN_STUFF,   /* mod_blosxom plugin stuff (NOT EDIT) */
    NULL,                   /* start function                      */
    mb_core_template,       /* template load function              */
    mb_core_entries,        /* find weblog entry function          */
    NULL,                   /* filter entry function               */
    NULL,                   /* end function                        */
    NULL,                   /* skip entry function                 */
    mb_core_interpolate,    /* interpolate function                */
    NULL,                   /* head function                       */
    mb_core_sort,           /* sort entry function                 */
    NULL,                   /* date format function                */
    NULL,                   /* story function                      */
    NULL,                   /* foot function                       */
    NULL                    /* last function                       */
};
