#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "ap_config.h"

#include <fcntl.h>
#include <sys/types.h>
#include <unistd.h>

#include "blosxom.h"


static void module_init(server_rec *s, pool *p)
{
    ap_add_version_component(MOD_BLOSXOM_STRING_VERSION);
}

static void *create_dir_conf(pool *p, char *path)
{
    blosxom_conf_rec *conf;

    conf = ap_pcalloc(p, sizeof(blosxom_conf_rec));
    conf->title = MB_BLOG_TITLE;
    conf->description = MB_BLOG_DESCRIPTION;
    conf->language = MB_BLOG_LANGUAGE;
    conf->datadir = NULL;
    conf->url = NULL;
    conf->depth = MB_DEPTH;
    conf->num_entries = MB_NUM_ENTRIES;
    conf->file_extension = MB_FILE_EXTENSION;
    conf->default_flavour = MB_DEFAULT_FLAVOUR;
    conf->plugin_dir = MB_PLUGIN_DIR;
    conf->plugin_state_dir = MB_PLUGIN_STATE_DIR;
    conf->plugin_name = ap_make_array(p, 5, sizeof(char *));
    return (void *)conf;
}


static int load_blosxom_plugin(blosxom_context_rec *ctx)
{
    blosxom_conf_rec *conf;
    int i;
    ap_os_dso_handle_t dso;

    conf = ap_get_module_config(ctx->request_rec->per_dir_config, &blosxom_module);
    blosxom_top_plugin = &core_blosxom_plugin;

    for (i = 0; i < conf->plugin_name->nelts; i++) {
        char *name, *path;
        blosxom_plugin *new_plugin;

        name = ap_pstrcat(ctx->pool, ((char **)conf->plugin_name->elts)[i], "_blosxom_plugin", NULL);
        path = ap_pstrcat(ctx->pool, conf->plugin_dir, "/", name, ".so", NULL);
        dso = ap_os_dso_load(ap_server_root_relative(ctx->pool, path));
        if (dso == NULL) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, ctx->request_rec,
                          "cannot load blosxom plugin '%s': %s",
                          path, ap_os_dso_error());
            continue;
        }
        new_plugin = ap_os_dso_sym(dso, name);
        if (new_plugin == NULL) {
            ap_log_rerror(APLOG_MARK, APLOG_ERR, ctx->request_rec,
                          "cannot dynamic link to '%s': %s",
                          name, ap_os_dso_error());
            ap_os_dso_unload(dso);
            continue;
        }

	new_plugin->prev = blosxom_top_plugin;
	blosxom_top_plugin = new_plugin;
    }

    return OK;
}

static int do_blosxom_plugin_start(blosxom_context_rec *ctx)
{
    blosxom_plugin *plugin = ctx->get_top_plugin();

    for (; plugin != NULL; plugin = plugin->prev) {
        if (plugin->start == NULL || plugin->start(ctx))
            continue;
        plugin->enable = 0;
    }

    return OK;
}

static int do_blosxom_plugin_template(blosxom_context_rec *ctx)
{
    blosxom_plugin *plugin = ctx->get_top_plugin();
    for (; plugin != NULL; plugin = plugin->prev) {
        if (plugin->template == NULL)
            continue;
        ctx->get_template = plugin->template; 
        return OK;
    }
    /* template plugin not found!? */
    ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, ctx->request_rec,
                  "template function is missing in blosxom plugins");
    return SERVER_ERROR;
}

static int do_blosxom_plugin_entries(blosxom_context_rec *ctx)
{
    blosxom_plugin *plugin = ctx->get_top_plugin();
    for (; plugin != NULL; plugin = plugin->prev) {
        int rc;

        if (plugin->entries == NULL)
            continue;
        rc = plugin->entries(ctx);
        if (rc == DECLINED)
            continue;
        return rc;
    }
    /* entries plugin not found!? */
    ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, ctx->request_rec,
                  "entries function is missing in blosxom plugins");
    return SERVER_ERROR;
}

static int do_blosxom_plugin_filter(blosxom_context_rec *ctx)
{
    blosxom_plugin *plugin = ctx->get_top_plugin();
    for (; plugin != NULL; plugin = plugin->prev) {
        if (plugin->filter == NULL)
            continue;
        plugin->filter(ctx);
    }

    return OK;
}

static int do_blosxom_plugin_skip(blosxom_context_rec *ctx)
{
    blosxom_plugin *plugin = ctx->get_top_plugin();
    for (; plugin != NULL; plugin = plugin->prev) {
        int rc;
        if (plugin->skip == NULL)
            continue;
        rc = plugin->skip(ctx);
        if (rc == DECLINED)
            continue;
        return rc;
    }

    return OK;
}

static int do_blosxom_plugin_interpolate(blosxom_context_rec *ctx)
{
    blosxom_plugin *plugin = ctx->get_top_plugin();
    for (; plugin != NULL; plugin = plugin->prev) {
        if (plugin->interpolate == NULL)
            continue;
        ctx->interpolate = plugin->interpolate;
	return OK;
    }

    /* interpolate plugin not found!? */
    ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, ctx->request_rec,
                  "interpolate function is missing in blosxom plugins");
    return SERVER_ERROR;
}

static int do_blosxom_plugin_head(blosxom_context_rec *ctx)
{
    blosxom_plugin *plugin = ctx->get_top_plugin();
    for (; plugin != NULL; plugin = plugin->prev) {
        if (plugin->head == NULL)
            continue;
    }

    return OK;
}

static int do_blosxom_plugin_sort(blosxom_context_rec *ctx)
{
    blosxom_plugin *plugin = ctx->get_top_plugin();
    for (; plugin != NULL; plugin = plugin->prev) {
        if (plugin->sort == NULL)
            continue;
        plugin->sort(ctx);
        return OK;
    }

    /* sort plugin not found!? */
    ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, ctx->request_rec,
                  "sort function is missing in blosxom plugins");
    return SERVER_ERROR;
}

static void set_common_bind_string(blosxom_context_rec *ctx)
{
    blosxom_conf_rec *conf;
    table *b = ctx->bind_string;

    conf = ap_get_module_config(ctx->request_rec->per_dir_config,
                                &blosxom_module);
    ap_table_set(b, "blog_title", conf->title);
    ap_table_set(b, "blog_description", conf->description);
    ap_table_set(b, "blog_language", conf->language);
    ap_table_set(b, "url", conf->url);
}

static void set_date_bind_string(blosxom_context_rec *ctx, blosxom_entry *ent)
{
    const struct tm *tm;
    table *b = ctx->bind_string;

    tm = localtime(&(ent->mtime));
    ap_table_set(b, "dw", ap_pstrdup(ctx->pool, ap_day_snames[tm->tm_wday]));
    ap_table_set(b, "mo", ap_pstrdup(ctx->pool, ap_month_snames[tm->tm_mon]));
    ap_table_set(b, "mo_num", ap_psprintf(ctx->pool, "%.2d", tm->tm_mon + 1));
    ap_table_set(b, "da", ap_psprintf(ctx->pool, "%.2d", tm->tm_mday));
    ap_table_set(b, "hr", ap_psprintf(ctx->pool, "%.2d", tm->tm_hour));
    ap_table_set(b, "min", ap_psprintf(ctx->pool, "%.2d", tm->tm_min));
    ap_table_set(b, "hr12", ap_psprintf(ctx->pool, "%d",
                                        tm->tm_hour >= 12
                                        ? (tm->tm_hour - 12) : (tm->tm_hour)));
    ap_table_set(b, "ampm", ap_pstrdup(ctx->pool,
                                       tm->tm_hour >= 12 ? "pm" : "am"));
    ap_table_set(b, "ti", ap_psprintf(ctx->pool, "%.2d:%.2d:%.2d",
                                      tm->tm_hour, tm->tm_min, tm->tm_sec));
    ap_table_set(b, "yr", ap_psprintf(ctx->pool, "%d", tm->tm_year + 1900));
		    
}

static int blosxom_handler(request_rec *r)
{
    blosxom_conf_rec *conf;
    blosxom_context_rec *ctx;
    blosxom_response_body body;
    int i;
    char *cur_date = "";
    char io_buff[HUGE_STRING_LEN];
    const char *foot;

    conf = ap_get_module_config(r->per_dir_config, &blosxom_module);

    /* parse request uri */
    ctx = mb_create_context(r);
    load_blosxom_plugin(ctx);
    do_blosxom_plugin_start(ctx);
    do_blosxom_plugin_template(ctx);
    do_blosxom_plugin_entries(ctx);
    do_blosxom_plugin_filter(ctx);
    if (do_blosxom_plugin_skip(ctx) != OK) {
        /* skip? */	    
    }
    do_blosxom_plugin_interpolate(ctx);
    do_blosxom_plugin_head(ctx);
    set_common_bind_string(ctx);
    do_blosxom_plugin_sort(ctx);

    r->content_type = ctx->get_template(ctx, "content_type");
    ap_send_http_header(r);
    if (r->header_only) {
        return OK;
    }

    body.head = ctx->interpolate(ctx, ctx->get_template(ctx, "head"));
    ap_rputs(body.head, r);
    for (i = 0; i< ctx->entries->nelts; i++) {
        const char *body_date, *story;
        blosxom_entry *ent = ((blosxom_entry **)ctx->entries->elts)[i];
        const struct tm *tm = localtime(&(ent->mtime));
        FILE *fp;

        if (conf->num_entries <= i
            && ctx->year == NULL && ctx->month == NULL && ctx->date == NULL)
        {
            break;
        }
        if (ctx->year != NULL && atoi(ctx->year) != tm->tm_year + 1900)
            continue;
        if (ctx->year != NULL && tm->tm_year + 1900 < atoi(ctx->year))
            break;
        if (ctx->num != NULL && tm->tm_mon + 1 != atoi(ctx->num))
            continue;
        if (ctx->date != NULL && atoi(ctx->date) != tm->tm_mday)
            continue;
        if (ctx->date != NULL && tm->tm_mday < atoi(ctx->date))
            break;

        body_date = ctx->get_template(ctx, "date");
        // do_blosxom_plugin_date(ctx, body);
        body_date = ctx->interpolate(ctx, body_date);
        if (strcmp(cur_date, body_date) != 0) {
            cur_date = ap_pstrdup(r->pool, body_date);
            ap_rputs(body_date, r);
        }

        if ((fp = ap_pfopen(r->pool, (char *)ent->source, "r")) != NULL) {
            char *title, *body = "";
            fgets(io_buff, sizeof(io_buff), fp);
            title = ap_pstrdup(r->pool, io_buff);
            title[strlen(title) - 1] = '\0';
            ap_table_set(ctx->bind_string, "title", title);

            while (fgets(io_buff, sizeof(io_buff), fp) != NULL) {
                body = ap_pstrcat(r->pool, body, io_buff, NULL);
            }
            ap_table_set(ctx->bind_string, "body", body);
            ap_table_set(ctx->bind_string, "raw",
                         ap_pstrcat(r->pool, title, "\n", body, NULL));
            ap_pfclose(r->pool, fp);
        }
        story = ctx->get_template(ctx, "story");
        // do_blosxom_plugin_story(ctx, ent);
        set_date_bind_string(ctx, ent);
        story = ctx->interpolate(ctx, story);
        ap_rputs(story, r);
    }

    foot = ctx->get_template(ctx, "foot");
    // do_blosxom_plugin_foot(ctx);
    foot = ctx->interpolate(ctx, foot);
    ap_rputs(foot, r);

    // do_blosxom_plugin_last(ctx);

    // do_blosxom_plugin_end(ctx);

    return OK;
}

/* Dispatch list of content handlers */
static const handler_rec handlers[] = { 
    { "blosxom", blosxom_handler }, 
    { NULL, NULL }
};

static const char *cmd_set_depth(cmd_parms *cmd, blosxom_conf_rec *m, char *depth)
{
    return NULL;
}

static const char *cmd_set_num_entries(cmd_parms *cmd, blosxom_conf_rec *m,
                                      char *str)
{
    int max;

    max = atoi(str);
    m->num_entries = max;

    return NULL;
}

static const char *cmd_add_plugin_list(cmd_parms *cmd, blosxom_conf_rec *m,
                                       char *name)
{
    *(char **)ap_push_array(m->plugin_name) = name;

    return NULL;
}

static const command_rec blosxom_module_cmds[] = {
    { "BlosxomTitle", ap_set_string_slot,
      (void *)XtOffsetOf(blosxom_conf_rec, title), OR_FILEINFO, TAKE1,
      "What's this blog's title?" },
    { "BlosxomDescription", ap_set_string_slot,
      (void *)XtOffsetOf(blosxom_conf_rec, description), OR_FILEINFO, TAKE1,
      "What's this blogs's description (for outgoing RSS feed)?" },
    { "BlosxomLanguage", ap_set_string_slot,
      (void *)XtOffsetOf(blosxom_conf_rec, language), OR_FILEINFO, TAKE1,
      "What's this blog's primary language (for outgoing RSS feed)?" },
    { "BlosxomDataDir", ap_set_file_slot,
      (void *)XtOffsetOf(blosxom_conf_rec, datadir), OR_FILEINFO, TAKE1,
      "Where are this blog's entries kept?" },
    { "BlosxomURL", ap_set_string_slot,
      (void *)XtOffsetOf(blosxom_conf_rec, url), OR_FILEINFO, TAKE1,
      "What's my preferred base URL for this blog (leave blank for automatic)?" },
    { "BlosxomDepth", cmd_set_depth, NULL, OR_FILEINFO, TAKE1,
      "0 = infinity depth, 1 = datadir only, n = n levels down" },
    { "BlosxomNumEntries", cmd_set_num_entries, NULL, OR_FILEINFO, TAKE1,
      "How many entries should I show on the home page?" },
    { "BlosxomFileExtension", ap_set_string_slot,
      (void *)XtOffsetOf(blosxom_conf_rec, file_extension), OR_FILEINFO, TAKE1,
      "What file extension signifies a mod_blosxom entry?" },
    { "BlosxomDefaultFlavour", ap_set_string_slot,
      (void *)XtOffsetOf(blosxom_conf_rec, default_flavour), OR_FILEINFO, TAKE1,
      "What is the default flavour?" },
    { "BlosxomPluginDir", ap_set_file_slot,
      (void *)XtOffsetOf(blosxom_conf_rec, plugin_dir), OR_FILEINFO, TAKE1,
      "Where are my plugins kept?" },
    { "BlosxomPluginStateDir", ap_set_file_slot,
      (void *)XtOffsetOf(blosxom_conf_rec, plugin_state_dir), OR_FILEINFO, TAKE1,
      "Where should my modules keep their state information?" },
    { "BlosxomAddPlugin", cmd_add_plugin_list, NULL, OR_FILEINFO, ITERATE,
      "What plugin is used?" },
    { NULL }
};

/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT blosxom_module = {
    STANDARD_MODULE_STUFF, 
    module_init,           /* module initializer                  */
    create_dir_conf,       /* create per-dir    config structures */
    NULL,                  /* merge  per-dir    config structures */
    NULL,                  /* create per-server config structures */
    NULL,                  /* merge  per-server config structures */
    blosxom_module_cmds,   /* table of config file commands       */
    handlers,              /* [#8] MIME-typed-dispatched handlers */
    NULL,                  /* [#1] URI to filename translation    */
    NULL,                  /* [#4] validate user id from request  */
    NULL,                  /* [#5] check if the user is ok _here_ */
    NULL,                  /* [#3] check access by host address   */
    NULL,                  /* [#6] determine MIME type            */
    NULL,                  /* [#7] pre-run fixups                 */
    NULL,                  /* [#9] log a transaction              */
    NULL,                  /* [#2] header parser                  */
    NULL,                  /* child_init                          */
    NULL,                  /* child_exit                          */
    NULL                   /* [#0] post read-request              */
#ifdef EAPI
   ,NULL,                  /* EAPI: add_module                    */
    NULL,                  /* EAPI: remove_module                 */
    NULL,                  /* EAPI: rewrite_command               */
    NULL                   /* EAPI: new_connection                */
#endif
};

